# w.c.s. - web application for online forms
# Copyright (C) 2005-2018  Entr'ouvert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import sys
import traceback

import psycopg2
from django.core.management.base import BaseCommand, CommandError

from wcs import sql
from wcs.carddef import CardDef
from wcs.formdef import FormDef
from wcs.qommon.misc import localstrftime
from wcs.qommon.publisher import UnknownTenantError, get_publisher_class
from wcs.roles import Role
from wcs.users import User


class Command(BaseCommand):

    help = 'Setup postgresql connection parameters and migrate existing objects.'

    def add_arguments(self, parser):
        parser.add_argument('-d', '--domain', required=True)
        parser.add_argument('--database', required=True)
        parser.add_argument('--host')
        parser.add_argument('--port', type=int)
        parser.add_argument('--user')
        parser.add_argument('--password')

    def handle(self, **options):
        self.publisher = self.get_publisher(options['domain'])
        if self.publisher.has_postgresql_config():
            raise CommandError('tenant already using postgresql')

        self.setup_connection(**options)
        sql.get_connection(new=True)

        # create early tables
        sql.do_snapshots_table()
        sql.do_custom_views_table()

        self.store_roles()
        self.store_users()
        self.store_forms()
        self.publisher.write_cfg()
        self.publisher.initialize_sql()
        self.publisher.cleanup()

    def get_publisher(self, domain):
        publisher_class = get_publisher_class()
        publisher = publisher_class.create_publisher()
        try:
            publisher.set_tenant_by_hostname(domain)
        except UnknownTenantError:
            raise CommandError('unknown tenant')
        return publisher

    def setup_connection(self, **kwargs):
        options = {}
        for k in ['host', 'port', 'database', 'user', 'password']:
            if k in kwargs:
                options[k] = kwargs.get(k)
        self.publisher.cfg['postgresql'] = options

    def store_users(self):
        self.convert_objects('user', User, sql.SqlUser, skip_global_forms_table_update=True)
        sql.SqlUser.fix_sequences()

    def store_roles(self):
        self.convert_objects('role', Role, sql.Role)

    def convert_objects(self, object_name, object_class, object_sql_class, **kwargs):
        errors = []
        print('converting %ss' % object_name)
        getattr(sql, 'do_%s_table' % object_name)()
        count = object_class.count()
        for i, obj_id in enumerate(object_class.keys()):
            obj = object_class.get(obj_id)
            obj.__class__ = object_sql_class
            try:
                obj.store(**kwargs)
            except AssertionError:
                errors.append((obj, traceback.format_exc()))
            self.update_progress(100 * i / count)

        if errors:
            with open('error_%s.log' % object_name, 'w') as error_log:
                for obj, trace in errors:
                    error_log.write('obj_id %s\n' % obj.id)
                    error_log.write(trace)
                    error_log.write('-' * 80)
                    error_log.write('\n\n')
            print('There were some errors, see error_%s.log for details.' % object_name)

    def store_forms(self):
        errors = []
        for formdef in FormDef.select(ignore_migration=True) + CardDef.select(ignore_migration=True):
            print('converting %s' % formdef.name)
            sql.do_formdef_tables(formdef, rebuild_views=True, rebuild_global_views=True)
            data_class = formdef.data_class(mode='files')

            if formdef.url_name != formdef.internal_identifier:
                # handle maybe different name
                setattr(
                    sys.modules['formdef'],
                    '_wcs_%s' % formdef.internal_identifier.title(),
                    getattr(sys.modules['formdef'], formdef.data_class_name),
                )

            count = data_class.count()

            # load all objects a first time, to allow the migrate() code to be
            # run and the eventual changes properly saved.
            for id in data_class.keys():
                try:
                    formdata = data_class.get(id)
                except AttributeError:
                    continue
            delattr(sys.modules[formdef.xml_root_node], formdef.data_class_name)

            # once this is done, reload and store everything in postgresql
            sql_data_class = formdef.data_class(mode='sql')
            for i, id in enumerate(data_class.keys()):
                try:
                    formdata = data_class.get(id)
                except AttributeError:
                    errors.append((formdata, traceback.format_exc()))
                    continue
                formdata._formdef = formdef
                formdata._evolution = formdata.evolution
                formdata.__class__ = sql_data_class
                try:
                    formdata.store()
                except (AssertionError, psycopg2.DataError):
                    errors.append((formdata, traceback.format_exc()))
                self.update_progress(100 * i / count)
            sql_data_class.fix_sequences()

            if formdef.migrate():
                # run formdef migration as we're done and table name can be stored etc.
                formdef.store()

        if errors:
            with open('error_formdata.log', 'w') as error_log:
                for formdata, trace in errors:
                    error_log.write(
                        '%s %s - %s\n' % (formdata.formdef, formdata.id, localstrftime(formdata.receipt_time))
                    )
                    error_log.write(trace)
                    error_log.write('-' * 80)
                    error_log.write('\n\n')
            print('There were some errors, see error_formdata.log.')

    def update_progress(self, progress, num_columns=120):
        sys.stdout.write(
            '[%s] %s%%\r'
            % (('#' * int((num_columns - 10) * progress / 100)).ljust(num_columns - 15), progress)
        )
